/* WMRecord -- A general purpose audio recording utility for Linux systems.
 *
 * Copyright (c)1999 Malcolm Cowe <malk@bruhaha.demon.co.uk>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * wmrecord.c -- core application source code file.
 * Created 28th June 1999.
 * 1.0.5 update released October 4th 1999.
 */


#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/soundcard.h>
#include <X11/xpm.h>
#include "wmgeneral/wmgeneral.h"
#include "wmrecord.h"

/* Colour pixmap of the wmrecord control-panel. */
#include "wmrecord.xpm"

/* Global variables containing status information with regard to the
 * audio device.
 */
#define MAXTRACK 99
#define BASEFILE "/.wmrecord/rec"
#define BUFSIZE 4096

char *dspstr, *mixstr;
int source = 0;
int samplerate = 0;
int bitrate = 0;
int stereo = 0;
int lvol = 0;
int rvol = 0;
int tracknum = 1;
int recording = 0;
int playing = 0;

struct coord pos[] = {
  {5,33,12,11},   /* Record */
  {17,33,12,11},  /* Stop */
  {29,33,12,11},  /* Play */
  {23,48,12,11},  /* Previous Track */
  {35,48,12,11},  /* Next Track */
  {47,48,12,11},  /* Erase Track */
  {47,5,12,9},    /* Input Source */
  {47,14,12,10},  /* Line Quality */
  {47,24,12,10},  /* Mono/Stereo */
  {47,34,12,10}   /* Bit Rate */
};
struct coord btn[] = {
  {0,88,12,11},   /* Record */
  {12,88,12,11},  /* Stop */
  {24,88,12,11},  /* Play */
  {37,88,12,11},  /* Previous Track */
  {49,88,12,11},  /* Next Track */
  {61,88,12,11},  /* Erase Track */
  {64,0,12,9},    /* Input Source */
  {64,9,12,10},   /* Line Quality */
  {64,19,12,10},  /* Mono/Stereo */
  {64,29,12,10}   /* Bit Rate */
};
struct coord btp[] = {
  {0,100,12,11},  /* Record */
  {12,100,12,11}, /* Stop */
  {24,100,12,11}, /* Play */
  {37,100,12,11}, /* Previous Track */
  {49,100,12,11}, /* Next Track */
  {61,100,12,11}, /* Erase Track */
  {77,0,12,9},    /* Input Source */
  {77,9,12,10},   /* Line Quality */
  {77,19,12,10},  /* Mono/Stereo */
  {77,29,12,10}   /* Bit Rate */
};


void wmSetStereo(int mode) {


  if (mode)
    stereo ? (stereo = 0) : (stereo = 1);
  copyXPMArea(109+(stereo*8), 65, 5, 3, 6, 19);

  return;
}


/* Write the current settings as defaults to the user's resource file.
 */
void wmSaveDefaults() {

  struct stat st;
  struct passwd *pw;
  int uid = getuid();
  FILE *fp;
  char *fres;


  pw = getpwuid(uid);
  fres = (char *)malloc(sizeof(char)*(strlen(pw->pw_dir)+11));
  sprintf(fres, "%s/.wmrecord",pw->pw_dir);
  if (stat(fres, &st)<0)
    mkdir(fres, S_IRWXU);
  fres = (char *)realloc(fres, sizeof(char)*(strlen(pw->pw_dir)+22));
  sprintf(fres, "%s/.wmrecord/wmrecordrc",pw->pw_dir);

  if ((fp = fopen(fres, "w+")) != NULL) {
    fprintf(fp, "source:\t\t");
    switch (source) {
    case SOUND_MIXER_CD:
      fprintf(fp, "cd\n");
      break;
    case SOUND_MIXER_LINE:
      fprintf(fp, "line\n");
      break;
    default:
      fprintf(fp, "mic\n");
      break;
    }
    fprintf(fp, "dsp:\t\t%s\n", dspstr);
    fprintf(fp, "mixer:\t\t%s\n", mixstr);
    fprintf(fp, "samplerate:\t%d\n", samplerate);
    fprintf(fp, "bitrate:\t%d\n", bitrate);
    fprintf(fp, "stereo:\t\t%s\n", stereo ? "on" : "off");
    fprintf(fp, "track:\t\t%d\n", tracknum);
    fprintf(fp, "volume:\t\t%d %d\n", lvol, rvol);
  }

  fclose(fp);
  free(fres);

  return;
}


void wmSetSampleRate(int mode) {

  char *p;
  int j = 0;


  if (mode)
    switch (samplerate) {
    case 8000:
      samplerate = 11025;
      break;
    case 11025:
      samplerate = 22050;
      break;
    case 22050:
      samplerate = 44100;
      break;
    case 44100:
      samplerate = 8000;
      break;
    }

  p = (char *)malloc(sizeof(char)*6);
  sprintf (p, "%05d", samplerate);

  if (p[0] == '0') {
    copyXPMArea(66, 79, 5, 7, 13, 21);
    j++;
  }
  while (j < 5) {
    copyXPMArea((p[j]-'0')*6 + 1, 79, 5, 7, 7+(6*(j+1)), 21);
    j++;
  }
  free (p);

  wmSaveDefaults();

  return;
}


void wmSetBitRate(int mode) {


  if (mode)
    (bitrate == 8) ? (bitrate = 16) : (bitrate = 8);
  copyXPMArea(109+(bitrate-8), 71, 5, 3, 6, 24);

  wmSaveDefaults();

  return;
}


void wmSetVolume(int mode) {

  int mixfd;
  int mask;
  int volume;


  if (mode > 0) {
    if (lvol < 100)
      lvol++;
    if (rvol < 100)
      rvol++;
  }
  else if (mode < 0) {
    if (lvol >0)
      lvol--;
    if (rvol >0)
      rvol--;
  }

  /* Open the audio mixer device. */
  if ((mixfd = open(mixstr, O_RDWR)) == -1) {
    fprintf(stderr, "WMRECORD: Could not open audio mixer device.\n");
    return;
  }

  /* Check that the Recording Mixer device is available (either the
   * microphone or line input).
   */
  if (ioctl(mixfd, SOUND_MIXER_READ_RECMASK, &mask) == -1)
    fprintf(stderr, "WMRECORD: Could not query audio mixer device.\n");
  else
    if (mask && source) {
      /* Set the volume of the recording source. */
      volume = (rvol << 8)|lvol;
      ioctl(mixfd, MIXER_WRITE(source), &volume);
      wmSaveDefaults();
    }
  /* Close the mixer. */
  close (mixfd);

  return;
}


void wmSelectInput(int mode) {


  /* Set source equal either to mic, line or CD audio. */
  if (mode) {
    if (source == SOUND_MIXER_CD)
      source = SOUND_MIXER_LINE;
    else if (source == SOUND_MIXER_LINE)
      source = SOUND_MIXER_MIC;
    else
      source = SOUND_MIXER_CD;
  }

  if (source == SOUND_MIXER_LINE)
    copyXPMArea(90, 0, 14, 14, 28, 6);
  else if (source == SOUND_MIXER_MIC)
    copyXPMArea(90, 15, 14, 14, 28, 6);
  else
    copyXPMArea(105, 0, 14, 14, 28, 6);

  wmSaveDefaults();

  return;
}


void wmSetTrackNum(int inc) {

  char *p, *q;
  int j = 0;
  struct passwd *pd;
  int uid = getuid();
  int fd, datalen;


  if (inc < 0) {
    if (--tracknum < 1)
      tracknum = MAXTRACK;
  }
  else if (inc > 0) {
    if (++tracknum > 99)
      tracknum = 1;
  }

  p = (char *)malloc(sizeof(char)*3);

  sprintf (p, "%02d", tracknum);
  if (p[0] == '0') {
    copyXPMArea(66, 79, 5, 7, 6, 50);
    j++;
  }
  while (j < 2) {
    copyXPMArea((p[j]-'0')*6 + 1, 79, 5, 7, 6*(j+1), 50);
    j++;
  }
  free (p);

  /* Initialize the counter display to the length, in seconds, of the
   * recording.
   */ 
  pd = getpwuid(uid);
  q = (char *)malloc(sizeof(char)*(strlen(pd->pw_dir)+strlen(BASEFILE)+7));
  sprintf(q, "%s%s%02d.wav",pd->pw_dir, BASEFILE, tracknum);
  /* Open the source file, where the recording has been stored. */
  p = (char *)malloc(sizeof(char)*4);
  if ((fd=open(q, O_RDONLY|O_SYNC))==-1)
    sprintf(p, "000");
  else {
    datalen = wmGetWavSettings(fd);
    close(fd);
    j = (samplerate * (bitrate >> 3) * (stereo ? 2 : 1));
    sprintf(p, "%03d", datalen/j);
  }
  free(q);

  j = 0;
  while (j < 3) {
    copyXPMArea((p[j]-'0')*7 + 1, 66, 6, 9, 6 + (7*j), 7);
    j++;
  }
  free(p);

  wmSaveDefaults();

  return;
}


void wmEraseTrack() {

  char *q;
  struct passwd *p;
  int uid = getuid();
  int j;
  char cse[3];

  p = getpwuid(uid);
  q = (char *)malloc(sizeof(char)*(strlen(p->pw_dir)+strlen(BASEFILE)+7));
  sprintf(q, "%s%s%02d.wav",p->pw_dir, BASEFILE, tracknum);
  unlink(q);

  /* Initialize the counter display. */
  sprintf(cse, "000");
  j = 0;
  while (j < 3) {
    copyXPMArea((cse[j]-'0')*7 + 1, 66, 6, 9, 6 + (7*j), 7);
    j++;
  }
  RedrawWindow();

  free (q);
  return;
}


/* Initialize the audio device before recording or playback of
 * samples.
 */
void wmSetupAudio(int *fd) {

  int mixfd, dspfd;
  int mask;
  int volume;
  int src;


  /* Open the audio mixer device. */
  if ((mixfd = open(mixstr, O_RDWR)) == -1) {
    fprintf(stderr, "WMRECORD: Could not open audio mixer device.\n");
    *fd = -1;
    return;
  }
  /* Check that the Recording Mixer device is available (either the
   * microphone or line input).
   */
  if (ioctl(mixfd, SOUND_MIXER_READ_RECMASK, &mask) == -1)
    fprintf(stderr, "WMRECORD: Could not query audio mixer device.\n");
  else
    if (mask && source) {
      /* Set the recording source to the correct device. */
      src = 1<<source;
      ioctl(mixfd, SOUND_MIXER_WRITE_RECSRC, &src);
      /* Set the volume of the recording source. */
      volume = (rvol << 8)|lvol;
      ioctl(mixfd, MIXER_WRITE(source), &volume);
    }
  /* Close the mixer. */
  close (mixfd);

  /* Open the audio DSP device in order to set the rest of the
   * parameters for the recording device.
   */
  if ((dspfd = open(dspstr, O_RDWR)) == -1) {
    fprintf(stderr, "WMRECORD: Could not open audio DSP device.\n");
    *fd = -1;
    return;
  }

  /* Set the bit rate. */
  src = (bitrate == 16) ? AFMT_S16_LE : AFMT_U8;
  if (ioctl (dspfd, SNDCTL_DSP_SETFMT, &src) <0) {
    fprintf (stderr, "WMRECORD: Unable to set audio sample size.\n");
    /* Close the DSP. */
    close (dspfd);
    *fd = -1;
    return;
  }

  /* Set the recording source to mono or stereo, as appropriate. */
  src = stereo;
  if (ioctl(dspfd, SNDCTL_DSP_STEREO, &src) <0) {
    fprintf (stderr, "WMRECORD: Unable to set audio channels.\n");
    /* Close the DSP. */
    close (dspfd);
    *fd = -1;
    return;
  }

  /* Set the sample rate. */
  src = samplerate;
  if (ioctl(dspfd, SNDCTL_DSP_SPEED, &src) <0) {
    fprintf (stderr, "WMRECORD: Unable to set sample rate.\n");
    /* Close the DSP. */
    close (dspfd);
    *fd = -1;
    return;
  }

  /* Set the calling argument to the DSP file descriptor. */
  *fd = dspfd;

  return;
}


void wmSetWavHeader(int fd, wavhdr *hdr) {


  strncpy((*hdr).riffID, "RIFF", 4);
  strncpy((*hdr).wavID, "WAVE", 4);
  strncpy((*hdr).fmtID, "fmt ", 4);
  (*hdr).fmtLen = 16;
  (*hdr).fmtTag = 1;
  (*hdr).nChannels = stereo ? 2 : 1;
  (*hdr).sampleRate = samplerate;
  (*hdr).bitsPerSample = bitrate;
  (*hdr).nBlockAlign = (*hdr).nChannels * (bitrate / 8);
  (*hdr).avgBytesPerSec = samplerate*(*hdr).nBlockAlign;
  strncpy((*hdr).dataID, "data", 4);
  write(fd, hdr, sizeof(*hdr));

  return;
}


int wmGetWavSettings(int fd) {

  char t1[4];
  unsigned long l1;
  int found;
  short fmt;
  unsigned short ch, brate;
  unsigned long srate;


  /* First read in the RIFF identifier. If this is missing then the
   * file is not a valid WAVE file.
   */
  if (read(fd, t1, 4)<4) {
    fprintf (stderr, "WMRECORD: Could not read from sound file.\n");
    return -1;
  }
  if (strncmp(t1, "RIFF", 4)) {
    fprintf (stderr, "WMRECORD: not a valid WAV file.\n");
    return -1;
  }
  /* Advance the file pointer to the next relevant field. */
  lseek(fd, 4, SEEK_CUR);
  /* Read in the WAVE identifier. */
  if (read(fd, t1, 4)<4) {
    fprintf (stderr, "WMRECORD: Could not read from sound file.\n");
    return -1;
  }
  if (strncmp(t1, "WAVE", 4)) {
    fprintf (stderr, "WMRECORD: not a valid WAV file.\n");
    return -1;
  }

  /* Search through the file for the format chunk. If the end of the
   * file is reached without finding the chunk, then the file is not a 
   * valid WAVE file.
   */
  found = 0;
  while (!found) {
    if (read(fd, t1, 4)<4) {
      fprintf (stderr, "WMRECORD: Could not read from sound file.\n");
      return -1;
    }
    if (strncmp(t1, "fmt ", 4)) {
      /* Determine the length of the chunk found and skip to the next
       * chunk. The chunk length is always stored in the four bytes
       * following the chunk id.
       */
      if (read(fd, &l1, 4)<4) {
	fprintf (stderr, "WMRECORD: Could not read from sound file.\n");
	return -1;
      }
      lseek(fd, l1, SEEK_CUR);
    }
    else {
      /* This is the format chunk, which stores the playback settings
       * for the recording.
       */
      /* Skip the length field, since we don't really need it. */
      lseek(fd, 4, SEEK_CUR);
      /* Read in the format tag. If it has a value of 1, then there is 
       * no compression and we can attempt to play the file
       * back. Otherwise, return.
       */
      if (read(fd, &fmt, 2)<2) {
	fprintf (stderr, "WMRECORD: Could not read from format chunk.\n");
	return -1;
      }
      if (fmt != 1) {
	fprintf (stderr, "WMRECORD: Wave file contains compressed data."
		 " Unable to continue.\n");
	return -1;
      }
      /* Get the stereo mode. */
      if (read(fd, &ch, 2)<2) {
	fprintf (stderr, "WMRECORD: Could not read from format chunk.\n");
	return -1;
      }
      else {
	stereo = (ch == 2) ? 1 : 0;
	wmSetStereo(0);
	RedrawWindow();
      }
      /* Get the sample rate. */
      if (read(fd, &srate, 4)<4) {
	fprintf (stderr, "WMRECORD: Could not read from format chunk.\n");
	return -1;
      }
      else {
	samplerate = srate;
	wmSetSampleRate(0);
	RedrawWindow();
      }
      /* Get the bit rate. This is at the end of the format chunk. */
      lseek(fd, 6, SEEK_CUR);
      if (read(fd, &brate, 2)<2) {
	fprintf (stderr, "WMRECORD: Could not read from format chunk.\n");
	return -1;
      }
      else {
	bitrate = brate;
	wmSetBitRate(0);
	RedrawWindow();
      }

      found++;
    }
  }

  /* Search through the file for the data chunk. If the end of the
   * file is reached without finding the chunk, then the file is not a 
   * valid WAVE file.
   */
  found = 0;
  while (!found) {
    if (read(fd, t1, 4)<4) {
      fprintf (stderr, "WMRECORD: Could not read from sound file.\n");
      return -1;
    }
    if (strncmp(t1, "data", 4)) {
      /* Determine the length of the chunk found and skip to the next
       * chunk. The chunk length is always stored in the four bytes
       * following the chunk id.
       */
      if (read(fd, &l1, 4)<4) {
	fprintf (stderr, "WMRECORD: Could not read from sound file.\n");
	return -1;
      }
      lseek(fd, l1, SEEK_CUR);
    }
    else {
      /* This is the data chunk, which stores the recording.
       */
      /* Get the length field. */
      if (read(fd, &l1, 4)<4) {
	fprintf (stderr, "WMRECORD: Could not read from sound file.\n");
	return -1;
      }
      else return l1;
    }
  }

  return 0;
}


void wmRecord() {

  int numread = 0;
  int total = 0; /* Total number of bytes read in so far. */
  int ttotal1 = 0;
  int ttotal2 = 0;
  int fd, dspfd; /* File descriptors for the output file and DSP device. */
  struct passwd *pw;
  int uid = getuid();
  char *fres;
  char abuf[BUFSIZE]; /* Audio input buffer. */
  int se = 0;         /* Seconds counter. */
  char cse[3];        /* String representation of the seconds counter. */
  int i = 0;
  int j, rt;
  XEvent Event;
  wavhdr hdr;


  /* Set up the audio device before starting to record. */
  wmSetupAudio(&dspfd);

  rt = (samplerate * (bitrate >> 3) * (stereo ? 2 : 1));
  /* Initialize the counter display. */
  sprintf(cse, "%03d", se++);
  j = 0;
  while (j < 3) {
    copyXPMArea((cse[j]-'0')*7 + 1, 66, 6, 9, 6 + (7*j), 7);
    j++;
  }
  RedrawWindow();

  /* Open the destination file, where the recording is to be stored. */
  pw = getpwuid(uid);
  fres=(char *)malloc(sizeof(char)*(strlen(pw->pw_dir)+strlen(BASEFILE)+7));
  sprintf(fres, "%s%s%02d.wav",pw->pw_dir, BASEFILE, tracknum);
  if ((fd=open(fres,O_RDWR|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR))==-1) {
    fprintf(stderr, "WMRECORD: Could not open destination file.\n");
    free(fres);
    close (dspfd);
    return;
  }
  free(fres);

  wmSetWavHeader(fd, &hdr);

  /* Main recording loop. */
  while (recording) {
    /* Read a chunk of data in from the recording device, and write it
     * out to disk.
     */
    numread = read (dspfd, abuf, BUFSIZE);
    write (fd, abuf, numread);
    total+=numread;

    /* Update the counter dislay. */
    ttotal1 = total / rt;
    if (ttotal1 - ttotal2) {
      if (se > 999)
	se=0;
      sprintf(cse, "%03d", se++);
      j = 0;
      while (j < 3) {
	copyXPMArea((cse[j]-'0')*7 + 1, 66, 6, 9, 6 + (7*j), 7);
	j++;
      }
      RedrawWindow();
      ttotal2 = ttotal1;
    }

    /* Process X Events while recording -- This is used to check
     * whether or not the stop button has been pressed. Not as elegant 
     * as a nice, multi-threaded application, but just as effective,
     * especially as the development platform is still libc5-based.
     */
    while (XPending (display)) {
      XNextEvent (display, &Event);
      switch (Event.type) {
      case ButtonPress:
	i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
	if (i == 1) {
	  copyXPMArea (btp[1].x, btp[1].y, btp[1].w, btp[1].h,
		       pos[1].x, pos[1].y);
	  RedrawWindow ();
	}
	break;
      case ButtonRelease:
	if (i == 1) {
	  copyXPMArea (btn[0].x, btn[0].y, btn[0].w, btn[0].h,
		       pos[0].x, pos[0].y);
	  copyXPMArea (btn[1].x, btn[1].y, btn[1].w, btn[1].h,
		       pos[1].x, pos[1].y);
	  recording = 0;
	  /* Adjust the WAV header to reflect the overall length of
	   * the RIFF data, and the length of the data chunk.
	   * It assumes that integers are 32-bit (4-byte) quantities.
	   */
	  lseek(fd, 4, 0);
	  i = total + 36;
	  write(fd, &i, sizeof(i));
	  lseek(fd, 40, SEEK_SET);
	  write(fd, &total, sizeof(total));

	  RedrawWindow ();
	}
	break;
      }
    }
  }

  close(dspfd); 
  close(fd); 
  return;
}


void wmPlay() {

  int numread = 0;
  int total = 0; /* Total number of bytes read in so far. */
  int ttotal1 = 0;
  int ttotal2 = 0;
  int fd, dspfd; /* File descriptors for the output file and DSP device. */
  struct passwd *pw;
  int uid = getuid();
  char *fres;
  char abuf[BUFSIZE];    /* Audio input buffer. */
  int se = 0;    /* Seconds counter. */
  char cse[3];   /* String representation of the seconds counter. */
  int i = 0;
  int j, rt;
  int datalen;
  XEvent Event;


  /* Open the source file, where the recording has been stored. */
  pw = getpwuid(uid);
  fres=(char *)malloc(sizeof(char)*(strlen(pw->pw_dir)+strlen(BASEFILE)+7));
  sprintf(fres, "%s%s%02d.wav",pw->pw_dir, BASEFILE, tracknum);
  if ((fd=open(fres,O_RDONLY|O_SYNC))==-1) {
    fprintf(stderr, "WMRECORD: Could not open source file.\n");
    free(fres);
    copyXPMArea (btn[2].x, btn[2].y, btn[2].w, btn[2].h,
		 pos[2].x, pos[2].y);
    RedrawWindow();
    playing = 0;
    return;
  }
  free(fres);

  /* Extract the audio settings (sample rate, mono/stereo, etc) from
   * the WAV file.
   */
  datalen = wmGetWavSettings(fd);

  /* Set up the audio device (based on the values extracted from the
   * wave file) before starting to play. */
  wmSetupAudio(&dspfd);

  rt = (samplerate * (bitrate >> 3) * (stereo ? 2 : 1));
  /* Initialize the counter display. */
  sprintf(cse, "%03d", se++);
  j = 0;
  while (j < 3) {
    copyXPMArea((cse[j]-'0')*7 + 1, 66, 6, 9, 6 + (7*j), 7);
    j++;
  }
  RedrawWindow();

  /* Main playback loop. */
  while (playing) {
    /* Read a chunk of data in from the recording device, and write it
     * out to disk.
     */
    numread = read (fd, abuf, BUFSIZE);
    if (numread>0) {
      write (dspfd, abuf, numread);
      total+=numread;
    }
    else {
      copyXPMArea (btn[2].x, btn[2].y, btn[2].w, btn[2].h,
		   pos[2].x, pos[2].y);
      RedrawWindow();
      playing = 0;
    }

    /* Update the counter dislay. */
    ttotal1 = total / rt;
    if (ttotal1 - ttotal2) {
      if (se > 999)
	se=0;
      sprintf(cse, "%03d", se++);
      j = 0;
      while (j < 3) {
	copyXPMArea((cse[j]-'0')*7 + 1, 66, 6, 9, 6 + (7*j), 7);
	j++;
      }
      RedrawWindow();
      ttotal2 = ttotal1;
    }

    /* Check to see if playback has reached the end of the data chunk
     * in the wave file.
     */
    if (total >= datalen) {
      copyXPMArea (btn[2].x, btn[2].y, btn[2].w, btn[2].h,
		   pos[2].x, pos[2].y);
      RedrawWindow();
      playing = 0;
    }

    /* Process X Events while recording -- This is used to check
     * whether or not the stop button has been pressed. Not as elegant 
     * as a nice, multi-threaded application, but just as effective,
     * especially as the development platform is still libc5-based.
     */
    while (XPending (display)) {
      XNextEvent (display, &Event);
      switch (Event.type) {
      case ButtonPress:
	i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
	if (i == 1) {
	  copyXPMArea (btp[1].x, btp[1].y, btp[1].w, btp[1].h,
		       pos[1].x, pos[1].y);
	  RedrawWindow ();
	}
	break;
      case ButtonRelease:
	if (i == 1) {
	  copyXPMArea (btn[2].x, btn[2].y, btn[2].w, btn[2].h,
		       pos[2].x, pos[2].y);
	  copyXPMArea (btn[1].x, btn[1].y, btn[1].w, btn[1].h,
		       pos[1].x, pos[1].y);
	  playing = 0;
	  RedrawWindow ();
	}
	break;
      }
    }
  }

  close(dspfd); 
  close(fd); 
  return;
}


int main(int argc, char *argv[]) {

  char flg;
  char *fres, *ins, *label, *value, *value2;
  FILE *fp;
  int gotsrc, gotdsp, gotmixer, gotrate, gotbit, gotstereo;
  int gottrack, gotvolume;
  char mask_bits[64*64];
  int mask_width = 64;
  int mask_height = 64;
  XEvent Event;
  int i, j;
  struct passwd *pw;
  int uid = getuid();


  /* Set up the display area for wmrecord. */
  createXBMfromXPM (mask_bits, wmrecord_xpm, mask_width, mask_height);
  openXwindow (argc, argv, wmrecord_xpm, mask_bits, mask_width, mask_height);
  /* Read in the defaults from the user's resource file.
   */
  pw = getpwuid(uid);
  fres = (char *)malloc(sizeof(char)*(strlen(pw->pw_dir)+22));
  sprintf(fres, "%s/.wmrecord/wmrecordrc",pw->pw_dir);

  gotsrc = gotdsp = gotmixer = gotrate = 0;
  gotbit = gotstereo = gottrack = gotvolume = 0;
  if ((fp = fopen(fres, "r")) != NULL) {
    ins = (char *) malloc(sizeof(char)*129);
    label = (char *) malloc(sizeof(char)*129);
    value = (char *) malloc(sizeof(char)*129);
    value2 = (char *) malloc(sizeof(char)*129);

    while (fgets(ins, 128, fp)!=NULL) {
      sscanf(ins, "%s", label);
      if (!strcmp(label, "source:") && !gotsrc) {
	sscanf(ins, "%*s %s", value);
	gotsrc=1;
	/* Set source equal either to mic, line or CD audio. */
	if (!strcmp(value, "cd"))
	  source = SOUND_MIXER_CD;
	else if (!strcmp(value, "line"))
	  source = SOUND_MIXER_LINE;
	else
	  source = SOUND_MIXER_MIC;
	gotsrc=1;
      }
      else if (!strcmp(label, "dsp:") && !gotdsp) {
	dspstr = (char *) malloc(sizeof(char)*129);
	sscanf(ins, "%*s %s", dspstr);
	gotdsp=1;
      }
      else if (!strcmp(label, "mixer:") && !gotmixer) {
	mixstr = (char *) malloc(sizeof(char)*129);
	sscanf(ins, "%*s %s", mixstr);
	gotmixer=1;
      }
      else if (!strcmp(label, "samplerate:") && !gotrate) {
	sscanf(ins, "%*s %s", value);
	samplerate = atoi(value);
	gotrate=1;
      }
      else if (!strcmp(label, "bitrate:") && !gotbit) {
	sscanf(ins, "%*s %s", value);
	bitrate = atoi(value);
	gotbit=1;
      }
      else if (!strcmp(label, "stereo:") && !gotstereo) {
	sscanf(ins, "%*s %s", value);
	if (!strcmp(value, "on"))
	  stereo = 1;
	else
	  stereo = 0;
	gotstereo=1;
      }
      else if (!strcmp(label, "track:") && !gottrack) {
	sscanf(ins, "%*s %s", value);
	tracknum = atoi(value);
	gottrack=1;
      }
      else if (!strcmp(label, "volume:") && !gotvolume) {
	if (sscanf(ins, "%*s %s %s", value, value2) == 2) {
	  lvol = atoi(value);
	  rvol = atoi(value2);
	  gotvolume=1;
	}
      }
    }
    fclose (fp);
    free(value);
    free(value2);
    free(label);
    free(ins);
  }
  free (fres);

  /* Read in the command line options. The command line can be used to
   * specify the audio dsp and mixer devices for wmrecord, and override
   * the settings in the resource file.
   */
  while ((flg = getopt(argc, argv, "d:m:h")) != EOF) {
    switch (flg) {
    case 'd':
      /* The value of the parameter is stored in external variable,
       * optarg.
       */
      if (gotdsp)
	free(dspstr);
      dspstr = (char *) malloc(sizeof(char)*(strlen(optarg)+1));
      strcpy(dspstr, optarg);
      gotdsp=1;
      break;
    case 'm':
      /* The value of the parameter is stored in external variable,
       * optarg.
       */
      if (gotmixer)
	free(mixstr);
      mixstr = (char *) malloc(sizeof(char)*(strlen(optarg)+1));
      strcpy(mixstr, optarg);
      gotmixer=1;
      break;
    case 'h':
    default:
      /* Print Usage. */
      printf ("WMRecord v1.0.5\n Copyright (c)1999 Malcolm Cowe "
	      "<malk@bruhaha.demon.co.uk>.\n");
      printf ("Home:\thttp://www.bruhaha.demon.co.uk\n\n");
      printf ("usage: %s [ options... ]\n\n", argv[0]);
      printf ("Options:\n\n");
      printf ("\t-d <dsp>\tSet DSP audio device (default \"/dev/dsp\").\n");
      printf ("\t-m <mixer>\tSet Mixer device (default \"/dev/mixer\").\n");
      printf ("\t-h \t\tPrint this message.\n");
      return 0;
      break;
    }
  }

  /* If any of the variables has not already been set, supply a default.
   * Note that the audio devices are not initialized at this point so
   * as to reduce conflicts with other applications. The devices are
   * only set up when the user wishes to record or play back.
   */
  if (!gotdsp) {
    dspstr = (char *) malloc(sizeof(char)*9);
    strcpy(dspstr, "/dev/dsp");
  }
  if (!gotmixer) {
    mixstr = (char *) malloc(sizeof(char)*11);
    strcpy(mixstr, "/dev/mixer");
  }
  if (!gotsrc)
    source = SOUND_MIXER_MIC;
  if (samplerate <=0)
    samplerate = 11025;
  if ((bitrate != 8) && (bitrate != 16))
    bitrate = 8;
  if (lvol <=0)
    lvol = 50;
  if (rvol <=0)
    rvol = 50;

  /* Set up the appropriate areas of the panel as buttons.
   */
  i = 0;
  for (i = 0; i < 10; i++)
    AddMouseRegion (i, pos[i].x, pos[i].y,
		    pos[i].x+pos[i].w, pos[i].y+pos[i].h);

  /* Set all the display areas to their initial value. */
  wmSetTrackNum(0);
  wmSetStereo(0);
  wmSetSampleRate(0);
  wmSetBitRate(0);
  wmSelectInput(0);

  /* Enter the Event Loop, monitoring for button presses and screen
   * refreshes, etc.
   */
  while (1) {
    /* Process any pending events.
     */
    while (XPending (display)) {
      XNextEvent (display, &Event);
      switch (Event.type) {
      case Expose:
	RedrawWindow ();
	break;
      case ClientMessage:
	break;
      case DestroyNotify:
	XCloseDisplay (display);
	exit (0);
	break;
      case ButtonPress:
	i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);
	switch (i) {
	case 0: /* Record button. */
	  if (!recording && !playing) {
	    recording = 1;
	    copyXPMArea (btp[0].x, btp[0].y, btp[0].w, btp[0].h,
			 pos[0].x, pos[0].y);
	    wmRecord();
	  }
	  break;
	case 1: /* Stop button. */
	  copyXPMArea (btp[1].x, btp[1].y, btp[1].w, btp[1].h,
		       pos[1].x, pos[1].y);
	  recording = playing = 0;
	  break;
	case 2: /* Play button. */
	  if (!recording && !playing) {
	    playing = 1;
	    copyXPMArea (btp[2].x, btp[2].y, btp[2].w, btp[2].h,
			 pos[2].x, pos[2].y);
	    wmPlay();
	  }
	  break;
	case 3: /* Previous Track */
	  if (!recording && !playing) {
	    copyXPMArea (btp[3].x, btp[3].y, btp[3].w, btp[3].h,
			 pos[3].x, pos[3].y);
	  }
	  break;
	case 4: /* Next Track */
	  if (!recording && !playing) {
	    copyXPMArea (btp[4].x, btp[4].y, btp[4].w, btp[4].h,
			 pos[4].x, pos[4].y);
	  }
	  break;
	case 5: /* Erase Track */
	  if (!recording && !playing) {
	    copyXPMArea (btp[5].x, btp[5].y, btp[5].w, btp[5].h,
			 pos[5].x, pos[5].y);
	  }
	  break;
	case 6: /* Input Select */
	  if (!recording && !playing) {
	    copyXPMArea (btp[6].x, btp[6].y, btp[6].w, btp[6].h,
			 pos[6].x, pos[6].y);
	    RedrawWindow();
	  }
	  break;
	case 7: /* Line Quality */
	  if (!recording && !playing) {
	    copyXPMArea (btp[7].x, btp[7].y, btp[7].w, btp[7].h,
			 pos[7].x, pos[7].y);
	  }
	  break;
	case 8: /* Mono / Stereo */
	  if (!recording && !playing) {
	    copyXPMArea (btp[8].x, btp[8].y, btp[8].w, btp[8].h,
			 pos[8].x, pos[8].y);
	  }
	  break;
	case 9: /* Bit Rate */
	  if (!recording && !playing) {
	    copyXPMArea (btp[9].x, btp[9].y, btp[9].w, btp[9].h,
			 pos[9].x, pos[9].y);
	  }
	  break;
	}
	RedrawWindow ();
	break;
      case ButtonRelease:
	if (i==1)
	  for (j=0; j<3; j++)
	    copyXPMArea (btn[j].x, btn[j].y, btn[j].w, btn[j].h,
			 pos[j].x, pos[j].y);
	else if ((i>2) && !recording && !playing) {
	  copyXPMArea (btn[i].x, btn[i].y, btn[i].w, btn[i].h,
		       pos[i].x, pos[i].y);

	  switch (i) {
	  case 3: /* Previous Track / Volume Down */
	    if (Event.xbutton.button == 3)
	      /* Volume Down */
	      wmSetVolume(-1);
	    else
	      /* Previous Track */
	      wmSetTrackNum(-1);
	    break;
	  case 4: /* Next Track  / Volume Up */
	    if (Event.xbutton.button == 3)
	      /* Volume Up */
	      wmSetVolume(1);
	    else
	      /* Next Track */
	      wmSetTrackNum(1);
	    break;
	  case 5: /* Erase Track */
	    wmEraseTrack();
	    break;
	  case 6: /* Input Select */
	    wmSelectInput(1);
	    break;
	  case 7: /* Line Quality */
	    wmSetSampleRate(1);
	    break;
	  case 8: /* Mono / Stereo */
	    wmSetStereo(1);
	    break;
	  case 9: /* Bit Rate */
	    wmSetBitRate(1);
	    break;
	  }
	}
	RedrawWindow();
	break;
      }
    }
    usleep(50000);
  }

  free (mixstr);
  free (dspstr);

  return 0;
}
