/* WMRecord -- A general purpose audio recording utility for Linux systems.
 *
 * Copyright (c)1999 Malcolm Cowe <malk@bruhaha.demon.co.uk>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * wmrecord.h -- header file for wmrecord program.
 * Created 4th October 1999.
 */


/* Structure for containing the WAV header. This structure is used
 * primarily for writing WAV files, since there may be many different
 * chunks in a WAV file, in arbitrary order. This program only writes
 * the Format ("fmt ") and DATA ("data") chunks when recording, but
 * other WAV files may have different content.
 *
 * WMRecord may not be able to play back WAV files generated by other
 * programs, since the primary purpose of this implementation is for
 * export to other applications (such as conversion to MP3).
 *
 * This header was taken from the Audio File Formats FAQ by
 * Chris Bagwell, cbagwell@sprynet.com and some other places too.
 */
typedef struct {
  char           riffID[4];
  unsigned long  riffLen;
  char           wavID[4];
  char           fmtID[4];
  unsigned long  fmtLen;
  unsigned short fmtTag;
  unsigned short nChannels;
  unsigned long  sampleRate;
  unsigned long  avgBytesPerSec;
  unsigned short nBlockAlign;
  unsigned short bitsPerSample;
  char           dataID[4];
  unsigned long  dataLen;
} wavhdr;

/* Structure for containing coordinates of the buttons in the pixmap. */
struct coord {
  int x;
  int y;
  int w;
  int h;
};

void wmSetStereo(int);
void wmSaveDefaults();
void wmSetSampleRate(int);
void wmSetBitRate(int);
void wmSetVolume(int);
void wmSelectInput(int);
void wmSetTrackNum(int);
void wmEraseTrack();
void wmSetupAudio(int *);
void wmSetWavHeader(int, wavhdr *);
 int wmGetWavSettings(int);
void wmRecord();
void wmPlay();
